#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>

#include <unistd.h>
#include <netdb.h>

#include <sys/stat.h>
#include <sys/time.h>

#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/fcntl.h>

#define	VERSION	"2.1"

#if 0
#include <stdarg.h>
static void
#define DP(X)	do { debugprintf X; } while (0)
debugprintf(const char *s, ...)
{
  va_list list;

  fprintf(stderr, "[");
  va_start(list, s);
  vfprintf(stderr, s, list);
  va_end(list);
  fprintf(stderr, "]\n");
}
#else
#define DP(X)	do { ; } while (0)
#endif

static int	sock;
static int	throttle;
static int	maxblock	= BUFSIZ;
static int	alarm_seconds;
static int	alarm_count;
static int	alarm_timeout;
static int	alarm_signal;

static long	copycount_in, copycount_out, copycount;

static void	alarm_ticker(void);

#ifdef	NO_SSL
#include "md5.h"
#else
#include "rsa.h"       /* SSLeay stuff */
#include "crypto.h"
#include "x509.h"
#include "pem.h"
#include "ssl.h"
#include "err.h"

static SSL	*ssl;
static SSL_CTX	*ssl_ctx;
#endif

static void
ex(const char *s)
{
  alarm(0);
  alarm(0);
  perror(s);
  sleep(1);
  exit(3);
}

static void
exs(const char *s)
{
  alarm(0);
  alarm(0);
  perror(s);
  ERR_print_errors_fp(stderr);
  sleep(1);
  exit(3);
}

#ifndef	NO_SSL
static void
tino_ssl_close(void)
{
  SSL_free(ssl);

  SSL_CTX_free(ssl_ctx);
}

static void
tino_ssl_client(void)
{
  SSLeay_add_all_algorithms();
  SSLeay_add_ssl_algorithms();
  SSL_load_error_strings();

  ssl_ctx = SSL_CTX_new(SSLv23_client_method());
  if (!ssl_ctx)
    ex("SSL_CTX_new");
  
  ssl = SSL_new(ssl_ctx);
  if (!ssl)
    exs("SSL_new");

  SSL_set_fd(ssl, sock);

  if (SSL_connect(ssl)!=1)
    exs("SSL_connect");
}

static void
tino_ssl_server(char *cert)
{
  SSLeay_add_all_algorithms();
  SSLeay_add_ssl_algorithms();
  SSL_load_error_strings();

  ssl_ctx = SSL_CTX_new(SSLv23_server_method());
  if (!ssl_ctx)
    ex("SSL_CTX_new");
  
  if (SSL_CTX_use_certificate_file(ssl_ctx, cert, SSL_FILETYPE_PEM)==-1)
    exs("SSL_CTX_use_cert...");

  if (SSL_CTX_use_PrivateKey_file(ssl_ctx, cert, SSL_FILETYPE_PEM)==-1)
    exs("SSL_CTX_use_Priv...");
  
  ssl = SSL_new(ssl_ctx);
  if (!ssl)
    exs("SSL_new");

  SSL_set_fd(ssl, sock);

  if (SSL_accept(ssl)==-1)
    exs("SSL_accept");
}
#endif

static void *
re_alloc(void *buf, size_t len)
{
  if (!len)
    len	= 1;
  if ((buf=(buf ? realloc(buf, len) : malloc(len)))==NULL)
    ex("malloc");
  return buf;
}

static char *
str_alloc(const char *s)
{
  size_t	len;
  char		*buf;

  len	= strlen(s)+1;
  buf	= re_alloc(NULL, len);
  strncpy(buf, s, len);
  buf[len-1]	= 0;
  return buf;
}

#define	USE_SIGACTION

static void
alarmhandler(int sig)
{
#ifndef USE_SIGACTION
  if (signal(SIGALRM, alarmhandler))
    ex("signal");
#endif
  alarm(alarm_seconds);
  if (++alarm_signal>alarm_count)
    ex("alarm");
}

static void
alarm_init(int s, int c)
{
#ifdef USE_SIGACTION
  struct sigaction sa;

  memset(&sa, 0, sizeof sa);
  sa.sa_handler	= alarmhandler;

  if (sigaction(SIGALRM, &sa, NULL))
    ex("sigaction");
#else
  signal(SIGALRM, alarmhandler);
#endif

  alarm_seconds	= s;
  alarm_count	= c;
  alarm_timeout	= 0;
  alarm_signal	= 0;

  alarm(s);
}

static void
alarm_process(void)
{
  if (alarm_signal)
    {
      if (++alarm_timeout>alarm_count)
	ex("timeout");

      alarm_ticker();

      alarm_signal	= 0;
    }
}

static void
alarm_reset(void)
{
  alarm_timeout	= 0;
}

static void
writen(const char *buf, size_t len, int fo)
{
  static long		delta;
  static struct timeval	last;
  int			pos, put;

  alarm_reset();
  for (pos=0; pos<len; pos+=put)
    {
      size_t	max;

      max	= len-pos;
      if (max>maxblock)
	max	= maxblock;

#ifndef NO_SSL
      if (ssl && sock==fo)
	put	= SSL_write(ssl, buf+pos, max);
      else
#endif
	put	= write(fo, buf+pos, max);
      alarm_process();

      if (put>0)
	{
	  if (put>max)
	    ex("oops");
	  alarm_reset();
	  if (throttle)
	    {
	      struct timeval	now;
	      long		hold, elapsed;

	      /* This code is self synchronizing
	       * to any type of problem.
	       */
	      delta	+= put;
	      gettimeofday(&now, NULL);
	      elapsed	=  (now.tv_sec-last.tv_sec)*1000+(now.tv_usec-last.tv_usec)/1000;
	      last	=  now;
	      if (elapsed<0)
		elapsed	= 0;
	      if (elapsed>1000000)
		elapsed	= 1000000;
	      delta	-= elapsed*throttle;
	      if (delta<0)
		delta	=  0;
	      hold	=  delta*1000/throttle;
	      if (hold>10000)
		{
#if 0
		  fprintf(stderr, " t=%ld d=%ld n=%ld\r", hold/1000, delta, elapsed);
#endif
		  fflush(stderr);
		  usleep(hold);
		}
	    }
	  continue;
	}

      if (!put)
	ex("premature EOF on output");

      put	= 0;
      if (errno!=EINTR && errno!=EAGAIN)
	ex("output");
    }
}

static int
readn(int fd, void *buf, size_t max)
{
  alarm_reset();
  for (;;)
    {
      int	got;

#ifndef NO_SSL
      if (ssl && sock==fd)
	got	= SSL_read(ssl, buf, max);
      else
#endif
	got	= read(fd, buf, max);
      alarm_process();

      if (got>=0)
	{
	  alarm_reset();
	  return got;
	}

      if (errno!=EINTR && errno!=EAGAIN)
	ex("input");
    }
}

#ifdef NO_ZLIB
static int
docopy(int fi, int fo, int ignore)
{
  char		*buf;
  int		got;
  size_t	bs;

  bs	= maxblock;
  if (bs<BUFSIZ)
    bs	= BUFSIZ;
  buf	= re_alloc(NULL, bs);
  while ((got=readn(fi, buf, bs))!=0)
    {
      writen(buf, got, fo);
      copycount	+= got;
    }
  free(buf);
  return 0;
}
#else
#include "zlib.h"
static void *
dozalloc(void *ign, unsigned item, unsigned size)
{
  return re_alloc(NULL, (long)item*(long)size);
}
static void
dozfree(void *ign, void *p)
{
  free(p);
}
static int
docopy(int fi, int fo, int doinflate)
{
  char		*bi, *bo;
  size_t	bs;
  z_stream	str;
  int		init, mode, ret;

  bs	= maxblock;
  if (bs<BUFSIZ)
    bs	= BUFSIZ;
  bi	= re_alloc(NULL, bs);
  bo	= re_alloc(NULL, bs);

  str.zalloc	= dozalloc;
  str.zfree	= dozfree;
  str.opaque	= 0;

  str.avail_in	= 0;
  str.next_in	= bi;

  str.avail_out	= bs;
  str.next_out	= bo;

  init		= 1;
  mode		= Z_NO_FLUSH;

  do
    {
      if (!str.avail_in && mode==Z_NO_FLUSH)
	{
	  if ((str.avail_in=readn(fi, str.next_in=bi, bs))==0)
	    mode	= Z_FINISH;
	  copycount_in	+= str.avail_in;
	  copycount	+= str.avail_in;
	}

      if (init)
	{
	  init	= 0;

	  if (doinflate)
	    {
	      if (inflateInit(&str)!=Z_OK)
		ex("inflate");
	    }
	  else if (deflateInit(&str, 9)!=Z_OK)
	    ex("deflate");
	}

      ret	= (doinflate ? inflate : deflate)(&str, mode);
#if 0
      fprintf(stderr, "\nmode=%d ret=%d\n", mode, ret);
#endif

      if (!str.avail_out || mode!=Z_NO_FLUSH || ret==Z_STREAM_END)
	{
	  size_t	put;

	  put	= bs-str.avail_out;
	  if (put)
	    writen(bo, put, fo);
	  else if (mode!=Z_NO_FLUSH && ret==Z_BUF_ERROR && !str.avail_in)
	    ex("starved");
#if 0
	  fprintf(stderr, "put=%d\n", put);
#endif
	  copycount_out	+= put;

	  str.avail_out	= bs;
	  str.next_out	= bo;
	}
    }
  while (ret==Z_OK || ret==Z_BUF_ERROR);

  /* 2002-07-14: Memory leak resolved	*/
  if (!init && (doinflate ? inflateEnd : deflateEnd)(&str)!=Z_OK)
     ex("zlib err");

  free(bi);
  free(bo);

  return ret!=Z_STREAM_END;
}
#endif

static void
swrite(const char *s)
{
  size_t	len;

  len	= strlen(s);

  writen(s, len, sock);
  writen("\n", 1, sock);
}

static void
swrite_free(char *s)
{
  swrite(s);
  free(s);
}

static char *
sread(void)
{
  char		buf[1024];
  char		*out	= 0;
  size_t	len	= 0;
  int		got;

  while ((got=readn(sock, buf, sizeof buf))>0)
    {
      char	*tmp;

      out	= re_alloc(out, len+got+1);
      memcpy(out+len, buf, got);
      out[len+got]	= 0;
      if ((tmp=strchr(out+len, '\n'))!=0)
	{
	  *tmp	= 0;
	  return out;
	}
      len	+= got;
    }
  out		= re_alloc(out, len+1);
  out[len]	= 0;
  return out;
}

static int
sread_match(const char *s, int n)
{
  char	*t;
  int	res;

  if (!n)
    n	= strlen(s)+1;
  t	= sread();
  res	= strncmp(t, s, n);
  free(t);
  return res;
}

static char *
md5sum(int fd, unsigned long max)
{
  MD5_CTX	ctx;
  char		buf[BUFSIZ], digest[16];
  int		got, i;
  char		*out, *ptr;
  unsigned long	total;

  MD5_Init(&ctx);
  total	= 0;
  while (total<max && (got=readn(fd, buf, sizeof buf))>0)
    {
      unsigned long tmp;

      if (got>(tmp=max-total))
	got	= (int)tmp;
      MD5_Update(&ctx, buf, got);
      total	+= got;
    }
  MD5_Final(digest, &ctx);

  out	= re_alloc(NULL, 32+32);
  sprintf(out, "%lu ", total);
  ptr	= out+strlen(out);
  for (i=0; i<16; i++)
    {
      *ptr++	= "0123456789abcdef"[(digest[i]>>4)&0xf];
      *ptr++	= "0123456789abcdef"[(digest[i]>>0)&0xf];
    }
  *ptr	= 0;
  return out;
}
